/*
 * dhcpcd - DHCP client daemon -
 * Copyright (C) 1996 - 1997 Yoichi Hariguchi <yoichi@fore.com>
 * Copyright (C) January, 1998 Sergei Viznyuk <sv@phystech.com>
 *
 * dhcpcd is an RFC2131 and RFC1541 compliant DHCP client daemon.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <net/route.h>
#include <net/if.h>
#include <arpa/nameser.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <syslog.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <resolv.h>
#include <netdb.h>
#include "kversion.h"
#include "pathnames.h"
#include "client.h"
#include <sys/wait.h>

extern	int			dhcpSocket;
extern	int			prev_ip_addr;
extern	int			Window;
extern  int			SetDHCPDefaultRoutes;
extern	int			TestCase;
extern	int			DebugFlag;
extern	int			SetDomainName;
extern	int			SetHostName;
extern	int			ReplResolvConf;
extern	int			KeepSearchlist;
extern	int			ReplNISConf;
extern	int			ReplNTPConf;
extern	int			IfName_len,IfNameExt_len;
extern	char			*IfName,*IfNameExt,*Cfilename,*ConfigDir;
extern	char			**ProgramEnviron;
extern	unsigned char		ClientHwAddr[ETH_ALEN],*ClientID;
extern	struct in_addr		default_router;
extern	dhcpInterface		DhcpIface;
extern	dhcpOptions		DhcpOptions;
extern	const dhcpMessage	*DhcpMsgRecv;

int	arpInform();

char	hostinfo_file[128];
int	resolv_renamed=0; 
int	yp_renamed=0;
int	ntp_renamed=0;  

/* Note: Legths initialised to negative to allow us to distinguish between "empty" and "not set" */
char InitialHostName[HOSTNAME_MAX_LEN];
int InitialHostName_len=-1;
char InitialDomainName[HOSTNAME_MAX_LEN];
int InitialDomainName_len=-1;

/*****************************************************************************/
char *cleanmetas(cstr) /* this is to clean single quotes out of DHCP strings */
char *cstr;		/* replace single quotes with space */
{
  register char *c=cstr;
  do
    if ( *c == 39 ) *c = ' ';
  while ( *c++ );
  return cstr;
}
/*****************************************************************************/
void execute_on_change(prm)
char *prm;
{
#ifdef EMBED
  if ( vfork() == 0 )
#else
  if ( fork() == 0 )
#endif
    {
      char *argc[5],exec_on_change[128];
      if ( Cfilename )
	snprintf(exec_on_change,sizeof(exec_on_change),Cfilename);
      else
	snprintf(exec_on_change,sizeof(exec_on_change),EXEC_ON_CHANGE);
      argc[0]=exec_on_change;
      argc[1]=hostinfo_file;
      argc[2]=prm;
      if ( DebugFlag )
        argc[3]="-d";
      else
        argc[3]=NULL;
      argc[4]=NULL;
      if ( execve(exec_on_change,argc,ProgramEnviron) && errno != ENOENT )
	syslog(LOG_ERR,"error executing \"%s %s %s\": %m\n",
	exec_on_change,hostinfo_file,prm);
      exit(0);
    }
}
/*****************************************************************************/
unsigned long getgenmask(ip_in)		/* this is to guess genmask	*/
unsigned long ip_in;			/* from network address		*/
{
  unsigned long t,p=ntohl(ip_in);
  if ( IN_CLASSA(p) )
    t= ~IN_CLASSA_NET;
  else
    {
      if ( IN_CLASSB(p) )
	t= ~IN_CLASSB_NET;
      else
	{
	  if ( IN_CLASSC(p) )
	    t= ~IN_CLASSC_NET;
	  else
	    t=0;
	}
    }
  while ( t&p ) t>>=1;
  return htonl(~t);
}
/*****************************************************************************/
int setDefaultRoute(route_addr)
char *route_addr;
{
struct	rtentry		rtent;
struct	sockaddr_in	*p;

memset(&rtent,0,sizeof(struct rtentry));
p			=	(struct sockaddr_in *)&rtent.rt_dst;
p->sin_family		=	AF_INET;
p->sin_addr.s_addr	=	0;
p			=	(struct sockaddr_in *)&rtent.rt_gateway;
p->sin_family		=	AF_INET;
memcpy(&p->sin_addr.s_addr,route_addr,4);
p			=	(struct sockaddr_in *)&rtent.rt_genmask;
p->sin_family		=	AF_INET;
p->sin_addr.s_addr	=	0;
#ifdef OLD_LINUX_VERSION
  rtent.rt_dev		=	IfName;
#else
  rtent.rt_dev		=	IfNameExt;
#endif
rtent.rt_metric	        =	1;
rtent.rt_window		=	Window;
rtent.rt_flags	        =	RTF_UP|RTF_GATEWAY|(Window ? RTF_WINDOW : 0);
if ( ioctl(dhcpSocket,SIOCADDRT,&rtent) == -1 )
  {
    if ( errno == ENETUNREACH )    /* possibly gateway is over the bridge */
      {                            /* try adding a route to gateway first */
	memset(&rtent,0,sizeof(struct rtentry));
	p			=	(struct sockaddr_in *)&rtent.rt_dst;
	p->sin_family		=	AF_INET;
	memcpy(&p->sin_addr.s_addr,route_addr,4);
	p			=	(struct sockaddr_in *)&rtent.rt_gateway;
	p->sin_family		=	AF_INET;
	p->sin_addr.s_addr	=	0;
	p			=	(struct sockaddr_in *)&rtent.rt_genmask;
	p->sin_family		=	AF_INET;
	p->sin_addr.s_addr	=	0xffffffff;
#ifdef OLD_LINUX_VERSION
	rtent.rt_dev		=	IfName;
#else
	rtent.rt_dev		=	IfNameExt;
#endif
	rtent.rt_metric     =	  0;
	rtent.rt_flags      =	  RTF_UP|RTF_HOST;
	if ( ioctl(dhcpSocket,SIOCADDRT,&rtent) == 0 )
	  {
	    memset(&rtent,0,sizeof(struct rtentry));
	    p			=	(struct sockaddr_in *)&rtent.rt_dst;
	    p->sin_family	=	AF_INET;
	    p->sin_addr.s_addr	=	0;
	    p			=	(struct sockaddr_in *)&rtent.rt_gateway;
	    p->sin_family	=	AF_INET;
	    memcpy(&p->sin_addr.s_addr,route_addr,4);
	    p			=	(struct sockaddr_in *)&rtent.rt_genmask;
	    p->sin_family	=	AF_INET;
	    p->sin_addr.s_addr	=	0;
#ifdef OLD_LINUX_VERSION
	    rtent.rt_dev	=	IfName;
#else
	    rtent.rt_dev	=	IfNameExt;
#endif
	    rtent.rt_metric	=	1;
	    rtent.rt_window	=	Window;
	    rtent.rt_flags	=	RTF_UP|RTF_GATEWAY|(Window ? RTF_WINDOW : 0);
	    if ( ioctl(dhcpSocket,SIOCADDRT,&rtent) == -1 )
	      {
		syslog(LOG_ERR,"dhcpConfig: ioctl SIOCADDRT: %m\n");
		return -1;
	      }
	  }
      }
    else
      {
	syslog(LOG_ERR,"dhcpConfig: ioctl SIOCADDRT: %m\n");
	return -1;
      }
  }
return 0;
}
/*****************************************************************************/
int dhcpConfig()
{
  int i;
  FILE *f;
  char hostinfo_file_old[128];
  struct ifreq		ifr;
  struct rtentry	rtent;
#ifdef OLD_LINUX_VERSION
  struct sockaddr_pkt	sap;
#endif
  struct sockaddr_in	*p = (struct sockaddr_in *)&(ifr.ifr_addr);
  struct hostent *hp=NULL;
  char *dname=NULL;
  int dname_len=0;
  char *cmdline[20];
  char pid[20];
  char *NameserverList;
  char *NewSearchlist;
  int ret;

  if ( TestCase ) goto tsc;
  memset(&ifr,0,sizeof(struct ifreq));
#ifdef OLD_LINUX_VERSION
  memcpy(ifr.ifr_name,IfName,IfName_len);
#else
  memcpy(ifr.ifr_name,IfNameExt,IfNameExt_len);
#endif
  p->sin_family = AF_INET;
  p->sin_addr.s_addr = DhcpIface.ciaddr;
  if ( ioctl(dhcpSocket,SIOCSIFADDR,&ifr) == -1 )  /* setting IP address */
    {
      syslog(LOG_ERR,"dhcpConfig: ioctl SIOCSIFADDR: %m\n");
      return -1;
    }
  memcpy(&p->sin_addr.s_addr,DhcpOptions.val[subnetMask],4);
  if ( ioctl(dhcpSocket,SIOCSIFNETMASK,&ifr) == -1 )  /* setting netmask */
    {
      p->sin_addr.s_addr = 0xffffffff; /* try 255.255.255.255 */
      if ( ioctl(dhcpSocket,SIOCSIFNETMASK,&ifr) == -1 )
	{
	  syslog(LOG_ERR,"dhcpConfig: ioctl SIOCSIFNETMASK: %m\n");
	  return -1;
	}
    }
  memcpy(&p->sin_addr.s_addr,DhcpOptions.val[broadcastAddr],4);
  if ( ioctl(dhcpSocket,SIOCSIFBRDADDR,&ifr) == -1 ) /* setting broadcast address */
    syslog(LOG_ERR,"dhcpConfig: ioctl SIOCSIFBRDADDR: %m\n");

  /* setting local route - not needed on later kernels  */
#ifdef OLD_LINUX_VERSION
  memset(&rtent,0,sizeof(struct rtentry));
  p			=	(struct sockaddr_in *)&rtent.rt_dst;
  p->sin_family		=	AF_INET;
  memcpy(&p->sin_addr.s_addr,DhcpOptions.val[subnetMask],4);
  p->sin_addr.s_addr	&=	DhcpIface.ciaddr;
  p			=	(struct sockaddr_in *)&rtent.rt_gateway;
  p->sin_family		=	AF_INET;
  p->sin_addr.s_addr	=	0;
  p			=	(struct sockaddr_in *)&rtent.rt_genmask;
  p->sin_family		=	AF_INET;
  memcpy(&p->sin_addr.s_addr,DhcpOptions.val[subnetMask],4);
  rtent.rt_dev		=	IfName;
  rtent.rt_metric     	=	1;
  rtent.rt_flags      	=	RTF_UP;
  if ( ioctl(dhcpSocket,SIOCADDRT,&rtent) )
    syslog(LOG_ERR,"dhcpConfig: ioctl SIOCADDRT: %m\n");
#endif

  for (i=0;i<DhcpOptions.len[staticRoute];i+=8)
    {  /* setting static routes */
      struct sockaddr_in *dstp; 
      struct sockaddr_in *gwp; 
      struct sockaddr_in *mskp; 
      memset(&rtent,0,sizeof(struct rtentry));
      dstp		=	(struct sockaddr_in *)&rtent.rt_dst;
      dstp->sin_family	=	AF_INET;
      memcpy(&dstp->sin_addr.s_addr,((char *)DhcpOptions.val[staticRoute])+i,4);
      gwp		=	(struct sockaddr_in *)&rtent.rt_gateway;
      gwp->sin_family	=	AF_INET;
      memcpy(&gwp->sin_addr.s_addr,((char *)DhcpOptions.val[staticRoute])+i+4,4);
      mskp		=	(struct sockaddr_in *)&rtent.rt_genmask;
      mskp->sin_family	=	AF_INET;
      mskp->sin_addr.s_addr = getgenmask(dstp->sin_addr.s_addr);
      rtent.rt_flags	=	RTF_UP|RTF_GATEWAY;
      if ( mskp->sin_addr.s_addr == 0xffffffff ) rtent.rt_flags |= RTF_HOST;

#ifdef OLD_LINUX_VERSION
      rtent.rt_dev	      =	  IfName;
#else
      rtent.rt_dev	      =	  IfNameExt;
#endif
      rtent.rt_metric     =	  1;
      if ( ioctl(dhcpSocket,SIOCADDRT,&rtent) )
	syslog(LOG_ERR,"dhcpConfig: ioctl SIOCADDRT: %m\n");
    }

  if ( SetDHCPDefaultRoutes )
    {
      if ( DhcpOptions.len[routersOnSubnet] > 3 )
	for (i=0;i<DhcpOptions.len[routersOnSubnet];i+=4)
	  setDefaultRoute(DhcpOptions.val[routersOnSubnet]);
    }
  else
    if ( default_router.s_addr > 0 ) setDefaultRoute((char *)&(default_router.s_addr));

  /* rebind dhcpSocket after changing ip address to avoid problems with 2.0 kernels */
#ifdef OLD_LINUX_VERSION
  memset(&sap,0,sizeof(sap));
  sap.spkt_family = AF_INET;
  sap.spkt_protocol = htons(ETH_P_ALL);
  memcpy(sap.spkt_device,IfName,IfName_len);
  if ( bind(dhcpSocket,(void*)&sap,sizeof(struct sockaddr)) == -1 )
    syslog(LOG_ERR,"dhcpConfig: bind: %m\n");
#endif  

  arpInform();
  if ( DebugFlag )
    fprintf(stdout,"dhcpcd: your IP address = %u.%u.%u.%u\n",
    ((unsigned char *)&DhcpIface.ciaddr)[0],
    ((unsigned char *)&DhcpIface.ciaddr)[1],
    ((unsigned char *)&DhcpIface.ciaddr)[2],
    ((unsigned char *)&DhcpIface.ciaddr)[3]);
  if ( ReplResolvConf )
    {
      char *p, OldSearchlist[256];
      OldSearchlist[0] = '\0';
      int a = 0;

      /* we might want to read the original resolv.conf first */
      /* and save the searchlist to append it to the one from the server */
      if ( KeepSearchlist )
        {
        int n;

        f = fopen(RESOLV_CONF, "r"); 
        if ( f ) 
          {
          while ( fgets(OldSearchlist, 256, f) )  /* read one line */
            {
              n = strlen(OldSearchlist);
              if ( n )  /* cut off trailing newline */
                {
                  p = OldSearchlist + strlen(OldSearchlist) - 1;
                  if ( *p == '\n' )  *p = 0;
                } else break;
              /* remove comments: */
              p = strchr(OldSearchlist, '#');
              if ( p )  *p = 0;
              /* search for something: */
              p = strstr(OldSearchlist, "search");
              if ( p==OldSearchlist )  	/* only if found at beginning of line */
                break;
              OldSearchlist[0] = '\0';
            } /* end while */
          fclose (f);
	  /* haven't found anything */
	  if (!OldSearchlist[0]) KeepSearchlist = 0;
          } 
        else 
          {
            /* open resolv.conf has failed */
            KeepSearchlist = 0;
          } /* end if ( f ) */
        } /* end if ( KeepSearchlist ) */

      /* the next four lines are commented out because it's now done by /sbin/modify_resolvconf 
       * automatically when a new resolv.conf is written */
#if 0
      resolv_renamed=1+rename(RESOLV_CONF,""RESOLV_CONF".sv");
      f=fopen(RESOLV_CONF,"w");
      if ( f )
	{
	  int i;
#endif
#if 0
	  if ( DhcpOptions.len[nisDomainName] )
	    fprintf(f,"domain %s\n",(char *)DhcpOptions.val[nisDomainName]);
	  else
	    if ( DhcpOptions.len[domainName] )
	      fprintf(f,"domain %s\n",(char *)DhcpOptions.val[domainName]);
#endif
	  NameserverList = malloc(DhcpOptions.len[dns] * 4 + 1);
	  NameserverList[0] = 0;
	  for (i=0;i<DhcpOptions.len[dns];i+=4)
	    sprintf(NameserverList + strlen(NameserverList), "%u.%u.%u.%u ",
	    ((unsigned char *)DhcpOptions.val[dns])[i],
	    ((unsigned char *)DhcpOptions.val[dns])[i+1],
	    ((unsigned char *)DhcpOptions.val[dns])[i+2],
	    ((unsigned char *)DhcpOptions.val[dns])[i+3]);
#if 0
	  if ( DhcpOptions.len[nisDomainName] + DhcpOptions.len[domainName] )
	    {
	      fprintf(f,"search");
	      if ( DhcpOptions.len[nisDomainName] )
	        fprintf(f," %s",(char *)DhcpOptions.val[nisDomainName]);
	      if ( DhcpOptions.len[domainName] )
	        fprintf(f," %s",(char *)DhcpOptions.val[domainName]);
	      fprintf(f,"\n");
	    }
#endif
 	  if ( DhcpOptions.len[dnsSearchPath] )
	    NewSearchlist = malloc(strlen(OldSearchlist) + 2 + strlen((char *)DhcpOptions.val[dnsSearchPath]) );
	  else if ( DhcpOptions.len[domainName] )
	    NewSearchlist = malloc(strlen(OldSearchlist) + 2 + strlen((char *)DhcpOptions.val[domainName]) );
	  else
	    NewSearchlist = malloc(strlen(OldSearchlist) + 2);
	  NewSearchlist[0] = '\0';

        if ( KeepSearchlist ) {
          if ( DhcpOptions.len[domainName] 
		  		/* add the new domain name only if it's not already in the old list */
			  && ( strstr(OldSearchlist, (char *)DhcpOptions.val[domainName]) == NULL )) { 
	    sprintf(NewSearchlist,"%s %s", 
			    (char *)DhcpOptions.val[domainName], 
			    OldSearchlist+7 ); /* +7 to skip leading string "search " */
	  } else {
	    sprintf(NewSearchlist,"%s", OldSearchlist+7 );
	  }
	} else {
 	    if ( DhcpOptions.len[dnsSearchPath] )
 	      sprintf(NewSearchlist,"%s", (char *)DhcpOptions.val[dnsSearchPath]);
	    else if ( DhcpOptions.len[domainName] )
	      sprintf(NewSearchlist,"%s",(char *)DhcpOptions.val[domainName]);
	  }

	cmdline[a++] = MODIFY_RESOLVCONF;
	cmdline[a++] = "modify";
	cmdline[a++] = "-s";
	cmdline[a++] = PACKAGE;
	cmdline[a++] = "-p";
	cmdline[a++] = PACKAGE;
	cmdline[a++] = "-i";
	sprintf(pid, "%d", getpid());
	cmdline[a++] = pid;
	if (NewSearchlist[0]) {
	  cmdline[a++] = "-l";
	  cmdline[a++] = NewSearchlist;
	}
	cmdline[a++] = "-n";
	cmdline[a++] = NameserverList;
	/* FIXME: the script name shouldn't be mandatory, because 
	   in this case, we haven't one */
	cmdline[a++] = "-f";
	cmdline[a++] = MODIFY_RESOLVCONF;
	cmdline[a++] = "-e";
	cmdline[a++] = IfNameExt;

	if ( DebugFlag ) {
	  char *msg;
	  if ( (msg = malloc(a*64+1)) == NULL)
	    syslog(LOG_ERR,"dhcpConfig: malloc error for msg: %m\n");
	  else {
	    msg[0] = '\0';
	    for (i=1;i<a;i++) {
	      strncat(msg,cmdline[i],63);
	      strcat(msg," ");
	    }
	    strcat(msg,"-t \"...\"");
	    syslog(LOG_DEBUG, "dhcpConfig: modify_resolvconf args: %s", msg);
	    free(msg);
	  }
	}

	cmdline[a++] = "-t";
	cmdline[a++] = RESOLVCONF_INFO;
	cmdline[a++] = NULL;

	{
	  int child = fork();
	  if (child == 0)
	    {
	      execv(MODIFY_RESOLVCONF, cmdline);
	      _exit(127);
	    }
	  if (child > 0)
	    {
	      int err;
	      do
		err = waitpid(child, &ret, 0);
	      while (err == -1 && errno == EINTR);
	    }
	  else
	    ret = -1;
	}

	if ( DebugFlag )
	  syslog(LOG_DEBUG, "dhcpConfig: modify_resolvconf returns %d", WEXITSTATUS(ret));
	free(NewSearchlist);
	free(NameserverList);
#if 0
	fclose(f);
      }
       else
 	syslog(LOG_ERR,"dhcpConfig: fopen: %m\n");
#endif

   /* moved the next section of code from before to after we've created
    * resolv.conf. See below for explanation. <poeml@suse.de>
    * res_init() is normally called from within the first function of the
    * resolver which is called. Here, we want resolv.conf to be
    * reread. Otherwise, we won't be able to find out about our hostname,
    * because the resolver won't notice the change in resolv.conf */
      (void)res_init();
    }
  if ( ReplNISConf )
    {
      yp_renamed=1+rename(NIS_CONF,""NIS_CONF".sv");
      f=fopen(NIS_CONF,"w");
      if ( f )
	{
	  int i;
	  char *domain=NULL;
	  if ( DhcpOptions.len[nisDomainName] )
	    domain=(char *)DhcpOptions.val[nisDomainName];
	  else
	    domain=(char *)DhcpOptions.val[domainName];
	  for (i=0;i<DhcpOptions.len[nisServers];i+=4)
	    fprintf(f,"domain %s server %u.%u.%u.%u\n",(domain?domain:"localdomain"),
	    ((unsigned char *)DhcpOptions.val[nisServers])[i],
	    ((unsigned char *)DhcpOptions.val[nisServers])[i+1],
	    ((unsigned char *)DhcpOptions.val[nisServers])[i+2],
	    ((unsigned char *)DhcpOptions.val[nisServers])[i+3]);
	  if ( !DhcpOptions.len[nisServers] )
	    fprintf(f,"domain %s broadcast\n", (domain?domain:"localdomain"));
	  fclose(f);
	}
      else
	syslog(LOG_ERR,"dhcpConfig: fopen: %m\n");
    }
  if ( ReplNTPConf )
    {
      ntp_renamed=1+rename(NTP_CONF,""NTP_CONF".sv");
      f=fopen(NTP_CONF,"w");
      if ( f )
 	{
 	  int net, mask;
 	  memcpy(&mask,DhcpOptions.val[subnetMask],4);
 	  net = DhcpIface.ciaddr & mask;

 	  /* Note: Revise drift/log file names and stratum for local clock */
 	  fprintf(f,"restrict default noquery notrust nomodify\n");
 	  fprintf(f,"restrict 127.0.0.1\n");
 	  fprintf(f,"restrict %u.%u.%u.%u mask %u.%u.%u.%u\n",
 		  ((unsigned char *)&net)[0],
 		  ((unsigned char *)&net)[1],
 		  ((unsigned char *)&net)[2],
 		  ((unsigned char *)&net)[3],
 		  ((unsigned char *)&mask)[0],
 		  ((unsigned char *)&mask)[1],
 		  ((unsigned char *)&mask)[2],
 		  ((unsigned char *)&mask)[3]);
 	  if ( DhcpOptions.len[ntpServers]>=4 )
	    {
	      int i;
	      char addr[4*3+3*1+1];
	      for (i=0;i<DhcpOptions.len[ntpServers];i+=4)
		{
		  snprintf(addr,sizeof(addr),"%u.%u.%u.%u",
		  ((unsigned char *)DhcpOptions.val[ntpServers])[i],
		  ((unsigned char *)DhcpOptions.val[ntpServers])[i+1],
		  ((unsigned char *)DhcpOptions.val[ntpServers])[i+2],
		  ((unsigned char *)DhcpOptions.val[ntpServers])[i+3]);
		  fprintf(f,"restrict %s\nserver %s\n",addr,addr);
		}
 	    }
	  else
	    {		/* No servers found, use local clock */
	      fprintf(f, "fudge 127.127.1.0 stratum 3\n");
 	      fprintf(f, "server 127.127.1.0\n");
	    }
 	  fprintf(f, "driftfile /var/lib/ntp/drift/ntp.drift\n");
 	  fprintf(f, "logfile /var/log/ntp\n");
 	  fclose(f);
 	}
       else
 	syslog(LOG_ERR,"dhcpConfig: fopen: %m\n");
     }
  if ( SetHostName )
    {
      char *p;
      if ( ! DhcpOptions.len[hostName] )
	{
	  hp=gethostbyaddr((char *)&DhcpIface.ciaddr,
	  sizeof(DhcpIface.ciaddr),AF_INET);
	  if ( hp )
	    {
	      dname=hp->h_name;
	      while ( *dname > 32 )
#if 0
		if ( *dname == '.' )
		  break;
		else
#endif
		  dname++;
	      dname_len=dname-hp->h_name;
	      DhcpOptions.val[hostName]=(char *)malloc(dname_len+1);
	      DhcpOptions.len[hostName]=dname_len;
	      memcpy((char *)DhcpOptions.val[hostName],
	      hp->h_name,dname_len);
	      ((char *)DhcpOptions.val[hostName])[dname_len]=0;
	      DhcpOptions.num++;
	    }
	}
      if ( InitialHostName_len<0 && gethostname(InitialHostName,sizeof(InitialHostName))==0 )
	{
	  InitialHostName_len=strlen(InitialHostName);
	  if ( DebugFlag )
	    fprintf(stdout,"dhcpcd: orig hostname = %s\n",InitialHostName);
	}
      if ( DhcpOptions.len[hostName] )
        {
	  /* Strip the domain from the fully qualified domain name
	   *
	   * We never want to set the hostname to the FQDN. 
	   * This would happen when gethostbyaddr() has to be used (see above), or
	   * when the server supplies the FQDN (e.g. when configured with 
	   * "use-host-decl-names on;")
	   *
	   * We think that `hostname` should always give the short hostname,
	   * the rest is the job of the resolver. 
	   */
	  p = strchr(DhcpOptions.val[hostName], '.');
	  if ( p ) {
		  *p = 0;
		  DhcpOptions.len[hostName] = strlen(DhcpOptions.val[hostName]);
	  }
          sethostname(DhcpOptions.val[hostName],DhcpOptions.len[hostName]);
	  /* syslog(LOG_DEBUG,"hostname set to %s\n", DhcpOptions.val[hostName]); */ /* syslog is most likely not running yes */
	  if ( DebugFlag )
	    fprintf(stdout,"dhcpcd: your hostname = %s\n",
	    (char *)DhcpOptions.val[hostName]);
	}
    }
  if ( SetDomainName )
    {
      if ( InitialDomainName_len<0 && getdomainname(InitialDomainName,sizeof(InitialDomainName))==0 )
	{
	  InitialDomainName_len=strlen(InitialDomainName);
	  if ( DebugFlag )
	    fprintf(stdout,"dhcpcd: orig domainname = %s\n",InitialDomainName);
	}
      if ( DhcpOptions.len[nisDomainName] )
        {
          setdomainname(DhcpOptions.val[nisDomainName],
		      DhcpOptions.len[nisDomainName]);
	  if ( DebugFlag )
	    fprintf(stdout,"dhcpcd: your domainname = %s\n",
		(char *)DhcpOptions.val[nisDomainName]);
        }
      else
        {
#if 0
	  if ( ! DhcpOptions.len[domainName] )
	    {
	      if ( ! hp )
		hp=gethostbyaddr((char *)&DhcpIface.ciaddr,
		sizeof(DhcpIface.ciaddr),AF_INET);
	      if ( hp )
		{
		  dname=hp->h_name;
		  while ( *dname > 32 )
		    if ( *dname == '.' )
		      {
			dname++;
		        break;
		      }
		    else
		      dname++;
		  dname_len=strlen(dname);
		  if ( dname_len )
		    {
		      DhcpOptions.val[domainName]=(char *)malloc(dname_len+1);
		      DhcpOptions.len[domainName]=dname_len;
		      memcpy((char *)DhcpOptions.val[domainName],
		      dname,dname_len);
		      ((char *)DhcpOptions.val[domainName])[dname_len]=0;
		      DhcpOptions.num++;
		    }
		}
	    }
          if ( DhcpOptions.len[domainName] )
            {
	      setdomainname(DhcpOptions.val[domainName],
		DhcpOptions.len[domainName]);
	      if ( DebugFlag )
		fprintf(stdout,"dhcpcd: your domainname = %s\n",
		(char *)DhcpOptions.val[domainName]);
	    }
#endif
	}
    }
tsc:
  memset(DhcpIface.version,0,sizeof(DhcpIface.version));
  strncpy(DhcpIface.version,VERSION,sizeof(DhcpIface.version));
  snprintf(hostinfo_file_old,sizeof(hostinfo_file_old),DHCP_CACHE_FILE,ConfigDir,IfNameExt);
  i=open(hostinfo_file_old,O_WRONLY|O_CREAT|O_TRUNC,S_IRUSR+S_IWUSR);
  if ( i == -1 ||
      write(i,(char *)&DhcpIface,sizeof(dhcpInterface)) == -1 ||
      close(i) == -1 )
    syslog(LOG_ERR,"dhcpConfig: open/write/close: %m\n");
  snprintf(hostinfo_file,sizeof(hostinfo_file),DHCP_HOSTINFO,ConfigDir,IfNameExt);
  snprintf(hostinfo_file_old,sizeof(hostinfo_file_old),""DHCP_HOSTINFO".old",ConfigDir,IfNameExt);
  rename(hostinfo_file,hostinfo_file_old);
  f=fopen(hostinfo_file,"w");
  if ( f )
    {
      int b,c;
      memcpy(&b,DhcpOptions.val[subnetMask],4);
      c = DhcpIface.ciaddr & b;
      fprintf(f,"\
IPADDR=%u.%u.%u.%u\n\
NETMASK=%u.%u.%u.%u\n\
NETWORK=%u.%u.%u.%u\n\
BROADCAST=%u.%u.%u.%u\n",
((unsigned char *)&DhcpIface.ciaddr)[0],
((unsigned char *)&DhcpIface.ciaddr)[1],
((unsigned char *)&DhcpIface.ciaddr)[2],
((unsigned char *)&DhcpIface.ciaddr)[3],
((unsigned char *)DhcpOptions.val[subnetMask])[0],
((unsigned char *)DhcpOptions.val[subnetMask])[1],
((unsigned char *)DhcpOptions.val[subnetMask])[2],
((unsigned char *)DhcpOptions.val[subnetMask])[3],
((unsigned char *)&c)[0],
((unsigned char *)&c)[1],
((unsigned char *)&c)[2],
((unsigned char *)&c)[3],
((unsigned char *)DhcpOptions.val[broadcastAddr])[0],
((unsigned char *)DhcpOptions.val[broadcastAddr])[1],
((unsigned char *)DhcpOptions.val[broadcastAddr])[2],
((unsigned char *)DhcpOptions.val[broadcastAddr])[3]);
      if ( DhcpOptions.len[routersOnSubnet] > 3 )
	{
	  fprintf(f,"\
GATEWAY=%u.%u.%u.%u",
((unsigned char *)DhcpOptions.val[routersOnSubnet])[0],
((unsigned char *)DhcpOptions.val[routersOnSubnet])[1],
((unsigned char *)DhcpOptions.val[routersOnSubnet])[2],
((unsigned char *)DhcpOptions.val[routersOnSubnet])[3]);
	  for (i=4;i<DhcpOptions.len[routersOnSubnet];i+=4)
  	    fprintf(f,",%u.%u.%u.%u",
	    ((unsigned char *)DhcpOptions.val[routersOnSubnet])[i],
	    ((unsigned char *)DhcpOptions.val[routersOnSubnet])[1+i],
	    ((unsigned char *)DhcpOptions.val[routersOnSubnet])[2+i],
	    ((unsigned char *)DhcpOptions.val[routersOnSubnet])[3+i]);
	}
if ( DhcpOptions.len[staticRoute] )
  {
    fprintf(f,"\nROUTE=%u.%u.%u.%u,%u.%u.%u.%u",
    ((unsigned char *)DhcpOptions.val[staticRoute])[0],
    ((unsigned char *)DhcpOptions.val[staticRoute])[1],
    ((unsigned char *)DhcpOptions.val[staticRoute])[2],
    ((unsigned char *)DhcpOptions.val[staticRoute])[3],
    ((unsigned char *)DhcpOptions.val[staticRoute])[4],
    ((unsigned char *)DhcpOptions.val[staticRoute])[5],
    ((unsigned char *)DhcpOptions.val[staticRoute])[6],
    ((unsigned char *)DhcpOptions.val[staticRoute])[7]);
    for (i=8;i<DhcpOptions.len[staticRoute];i+=8)
    fprintf(f,",%u.%u.%u.%u,%u.%u.%u.%u",
    ((unsigned char *)DhcpOptions.val[staticRoute])[i],
    ((unsigned char *)DhcpOptions.val[staticRoute])[1+i],
    ((unsigned char *)DhcpOptions.val[staticRoute])[2+i],
    ((unsigned char *)DhcpOptions.val[staticRoute])[3+i],
    ((unsigned char *)DhcpOptions.val[staticRoute])[4+i],
    ((unsigned char *)DhcpOptions.val[staticRoute])[5+i],
    ((unsigned char *)DhcpOptions.val[staticRoute])[6+i],
    ((unsigned char *)DhcpOptions.val[staticRoute])[7+i]);
  }
if ( DhcpOptions.len[hostName] )
  fprintf(f,"\nHOSTNAME=\'%s\'",cleanmetas((char *)DhcpOptions.val[hostName]));
if ( DhcpOptions.len[domainName] )
  fprintf(f,"\nDOMAIN=\'%s\'",cleanmetas((char *)DhcpOptions.val[domainName]));
if ( DhcpOptions.len[nisDomainName] )
  fprintf(f,"\nNISDOMAIN=\'%s\'",cleanmetas((char *)DhcpOptions.val[nisDomainName]));
if ( DhcpOptions.len[rootPath] )
  fprintf(f,"\nROOTPATH=\'%s\'",cleanmetas((char *)DhcpOptions.val[rootPath]));
fprintf(f,"\n\
DNS=%u.%u.%u.%u",
((unsigned char *)DhcpOptions.val[dns])[0],
((unsigned char *)DhcpOptions.val[dns])[1],
((unsigned char *)DhcpOptions.val[dns])[2],
((unsigned char *)DhcpOptions.val[dns])[3]);
for (i=4;i<DhcpOptions.len[dns];i+=4)
  fprintf(f,",%u.%u.%u.%u",
  ((unsigned char *)DhcpOptions.val[dns])[i],
  ((unsigned char *)DhcpOptions.val[dns])[1+i],
  ((unsigned char *)DhcpOptions.val[dns])[2+i],
  ((unsigned char *)DhcpOptions.val[dns])[3+i]);
if ( DhcpOptions.len[dnsSearchPath] )
  fprintf(f, "\nDNSSEARCH=\'%s\'", cleanmetas((char *)DhcpOptions.val[dnsSearchPath]));
if ( DhcpOptions.len[ntpServers]>=4 )
  {
    fprintf(f,"\nNTPSERVERS=%u.%u.%u.%u",
    ((unsigned char *)DhcpOptions.val[ntpServers])[0],
    ((unsigned char *)DhcpOptions.val[ntpServers])[1],
    ((unsigned char *)DhcpOptions.val[ntpServers])[2],
    ((unsigned char *)DhcpOptions.val[ntpServers])[3]);
    for (i=4;i<DhcpOptions.len[ntpServers];i+=4)
      fprintf(f,",%u.%u.%u.%u",
      ((unsigned char *)DhcpOptions.val[ntpServers])[i],
      ((unsigned char *)DhcpOptions.val[ntpServers])[1+i],
      ((unsigned char *)DhcpOptions.val[ntpServers])[2+i],
      ((unsigned char *)DhcpOptions.val[ntpServers])[3+i]);
  }
if ( DhcpOptions.len[netBIOSnameServer]>=4 )
  {
    fprintf(f,"\nWINSSERVERS=%u.%u.%u.%u",
    ((unsigned char *)DhcpOptions.val[netBIOSnameServer])[0],
    ((unsigned char *)DhcpOptions.val[netBIOSnameServer])[1],
    ((unsigned char *)DhcpOptions.val[netBIOSnameServer])[2],
    ((unsigned char *)DhcpOptions.val[netBIOSnameServer])[3]);
    for (i=4;i<DhcpOptions.len[netBIOSnameServer];i+=4)
      fprintf(f,",%u.%u.%u.%u",
      ((unsigned char *)DhcpOptions.val[netBIOSnameServer])[i],
      ((unsigned char *)DhcpOptions.val[netBIOSnameServer])[1+i],
      ((unsigned char *)DhcpOptions.val[netBIOSnameServer])[2+i],
      ((unsigned char *)DhcpOptions.val[netBIOSnameServer])[3+i]);
  }
if ( DhcpOptions.len[netBIOSdgramDistServer]>=4 )
  {
    fprintf(f,"\nNETBIOSDGRAMDISTSERVER=%u.%u.%u.%u",
    ((unsigned char *)DhcpOptions.val[netBIOSdgramDistServer])[0],
    ((unsigned char *)DhcpOptions.val[netBIOSdgramDistServer])[1],
    ((unsigned char *)DhcpOptions.val[netBIOSdgramDistServer])[2],
    ((unsigned char *)DhcpOptions.val[netBIOSdgramDistServer])[3]);
    for (i=4;i<DhcpOptions.len[netBIOSdgramDistServer];i+=4)
      fprintf(f,",%u.%u.%u.%u",
      ((unsigned char *)DhcpOptions.val[netBIOSdgramDistServer])[i],
      ((unsigned char *)DhcpOptions.val[netBIOSdgramDistServer])[1+i],
      ((unsigned char *)DhcpOptions.val[netBIOSdgramDistServer])[2+i],
      ((unsigned char *)DhcpOptions.val[netBIOSdgramDistServer])[3+i]);
  }
if ( DhcpOptions.len[netBIOSnodeType]>0 )
  {
    fprintf(f,"\nNETBIOSNODETYPE=%u",
	 *(unsigned char *)DhcpOptions.val[netBIOSnodeType]);
  }
if ( DhcpOptions.len[netBIOSscope]>0 )
  {
	 fprintf(f,"\nNETBIOSSCOPE=%s",
	 (char *)DhcpOptions.val[netBIOSscope]);
  }
if ( DhcpOptions.len[nisServers]>=4 )
  {
    fprintf(f,"\nNISSERVERS=%u.%u.%u.%u",
    ((unsigned char *)DhcpOptions.val[nisServers])[0],
    ((unsigned char *)DhcpOptions.val[nisServers])[1],
    ((unsigned char *)DhcpOptions.val[nisServers])[2],
    ((unsigned char *)DhcpOptions.val[nisServers])[3]);
    for (i=4;i<DhcpOptions.len[nisServers];i+=4)
      fprintf(f,",%u.%u.%u.%u",
      ((unsigned char *)DhcpOptions.val[nisServers])[i],
      ((unsigned char *)DhcpOptions.val[nisServers])[1+i],
      ((unsigned char *)DhcpOptions.val[nisServers])[2+i],
      ((unsigned char *)DhcpOptions.val[nisServers])[3+i]);
  }
if ( DhcpOptions.len[lprServer] > 3 )
  {
    fprintf(f,"\
\nLPRSERVERS=%u.%u.%u.%u",
((unsigned char *)DhcpOptions.val[lprServer])[0],
((unsigned char *)DhcpOptions.val[lprServer])[1],
((unsigned char *)DhcpOptions.val[lprServer])[2],
((unsigned char *)DhcpOptions.val[lprServer])[3]);
	  for (i=4;i<DhcpOptions.len[lprServer];i+=4)
  	    fprintf(f,",%u.%u.%u.%u",
	    ((unsigned char *)DhcpOptions.val[lprServer])[i],
	    ((unsigned char *)DhcpOptions.val[lprServer])[1+i],
	    ((unsigned char *)DhcpOptions.val[lprServer])[2+i],
	    ((unsigned char *)DhcpOptions.val[lprServer])[3+i]);
	}

if ( DhcpOptions.len[logServer] > 3 )
  {
    fprintf(f,"\
\nLOGSERVERS=%u.%u.%u.%u",
((unsigned char *)DhcpOptions.val[logServer])[0],
((unsigned char *)DhcpOptions.val[logServer])[1],
((unsigned char *)DhcpOptions.val[logServer])[2],
((unsigned char *)DhcpOptions.val[logServer])[3]);
	  for (i=4;i<DhcpOptions.len[logServer];i+=4)
  	    fprintf(f,",%u.%u.%u.%u",
	    ((unsigned char *)DhcpOptions.val[logServer])[i],
	    ((unsigned char *)DhcpOptions.val[logServer])[1+i],
	    ((unsigned char *)DhcpOptions.val[logServer])[2+i],
	    ((unsigned char *)DhcpOptions.val[logServer])[3+i]);
	}

fprintf(f,"\n\
DHCPSID=%u.%u.%u.%u\n\
DHCPGIADDR=%u.%u.%u.%u\n\
DHCPSIADDR=%u.%u.%u.%u\n\
DHCPCHADDR=%02X:%02X:%02X:%02X:%02X:%02X\n\
DHCPSHADDR=%02X:%02X:%02X:%02X:%02X:%02X\n\
DHCPSNAME=\'%s\'\n\
LEASETIME=%u\n\
RENEWALTIME=%u\n\
REBINDTIME=%u\n\
INTERFACE=\'%s\'\n\
CLASSID=\'%s\'\n",
((unsigned char *)DhcpOptions.val[dhcpServerIdentifier])[0],
((unsigned char *)DhcpOptions.val[dhcpServerIdentifier])[1],
((unsigned char *)DhcpOptions.val[dhcpServerIdentifier])[2],
((unsigned char *)DhcpOptions.val[dhcpServerIdentifier])[3],
((unsigned char *)&DhcpMsgRecv->giaddr)[0],
((unsigned char *)&DhcpMsgRecv->giaddr)[1],
((unsigned char *)&DhcpMsgRecv->giaddr)[2],
((unsigned char *)&DhcpMsgRecv->giaddr)[3],
((unsigned char *)&DhcpMsgRecv->siaddr)[0],
((unsigned char *)&DhcpMsgRecv->siaddr)[1],
((unsigned char *)&DhcpMsgRecv->siaddr)[2],
((unsigned char *)&DhcpMsgRecv->siaddr)[3],
ClientHwAddr[0],
ClientHwAddr[1],
ClientHwAddr[2],
ClientHwAddr[3],
ClientHwAddr[4],
ClientHwAddr[5],
DhcpIface.shaddr[0],
DhcpIface.shaddr[1],
DhcpIface.shaddr[2],
DhcpIface.shaddr[3],
DhcpIface.shaddr[4],
DhcpIface.shaddr[5],
cleanmetas(DhcpMsgRecv->sname),
ntohl(*(unsigned int *)DhcpOptions.val[dhcpIPaddrLeaseTime]),
ntohl(*(unsigned int *)DhcpOptions.val[dhcpT1value]),
ntohl(*(unsigned int *)DhcpOptions.val[dhcpT2value]),
IfNameExt,
DhcpIface.class_id);
      if ( ClientID )
	fprintf(f,"CLIENTID=\'%s\'\n",ClientID);
      else
	fprintf(f,"CLIENTID=%02X:%02X:%02X:%02X:%02X:%02X\n",
DhcpIface.client_id[3],DhcpIface.client_id[4],DhcpIface.client_id[5],
DhcpIface.client_id[6],DhcpIface.client_id[7],DhcpIface.client_id[8]);
      fclose(f);
    }
  else
    syslog(LOG_ERR,"dhcpConfig: fopen: %m\n");
#if 0
  if ( Cfilename )
    if ( fork() == 0 )
      {
	char *argc[2];
	argc[0]=Cfilename;
	argc[1]=NULL;
	if ( execve(Cfilename,argc,ProgramEnviron) )
	  syslog(LOG_ERR,"error executing \"%s\": %m\n",
	  Cfilename);
	exit(0);
      }
#endif
  if ( DhcpIface.ciaddr == prev_ip_addr )
    execute_on_change("up");
  else					/* IP address has changed */
    {
      execute_on_change("new");
      prev_ip_addr=DhcpIface.ciaddr;
    }
  if ( *(unsigned int *)DhcpOptions.val[dhcpIPaddrLeaseTime] == 0xffffffff )
    {
      syslog(LOG_INFO,"infinite IP address lease time. Exiting\n");
      exit(0);
    }
  return 0;
}

/*****************************************************************************/
int dhcpConfigOptionSet(char *optionString, int optionIdx, char *sptr)
{
  int	i = 0;
  char  *eql;

  if( !optionString || !sptr || (optionIdx < 0) )
    return 0;
#ifdef DEBUG
  (void) fprintf(stderr,"dhcpConfigOptionSet(): optionString=%s,optionIdx=%i,sptr=\"%s\"\n",
		 optionString,optionIdx,sptr);
#endif

  eql = sptr;
  eql++;			/* skip over '=' char */
  for( i = 0; i < 4; i++ ) {
    ((unsigned char *)DhcpOptions.val[optionIdx])[i] = strtol(eql, (char **)NULL, 10);
    while( (i != 3) && eql && (*eql != '.') ) { eql++; }	 /* Skip over octet */
    eql++;			                         /* skip over period */
  }
  eql++;			/* skip over comma if exists */
  for (i = 4; i < DhcpOptions.len[optionIdx];i += 4) {
    ((unsigned char *)DhcpOptions.val[optionIdx])[i] = strtol(eql, (char **)NULL, 10);
    while( eql && (*eql != '.') ) { eql++; }	 /* Skip over octet */
    eql++;			                         /* skip over period */
    ((unsigned char *)DhcpOptions.val[optionIdx])[i + 1] = strtol(eql, (char **)NULL, 10);
    while( eql && (*eql != '.') ) { eql++; }	 /* Skip over octet */
    eql++;			                         /* skip over period */
    ((unsigned char *)DhcpOptions.val[optionIdx])[i + 2] = strtol(eql, (char **)NULL, 10);
    while( eql && (*eql != '.') ) { eql++; }	 /* Skip over octet */
    eql++;			                         /* skip over period */
    ((unsigned char *)DhcpOptions.val[optionIdx])[i + 3] = strtol(eql, (char **)NULL, 10);
    while( eql && (*eql != ',') ) { eql++; }	/* Skip over comma */
    eql++;
  }
#ifdef DEBUG
  (void) fprintf(stdout,"%s=%u.%u.%u.%u",optionString,
		 ((unsigned char *)DhcpOptions.val[optionIdx])[0],
		 ((unsigned char *)DhcpOptions.val[optionIdx])[1],
		 ((unsigned char *)DhcpOptions.val[optionIdx])[2],
		 ((unsigned char *)DhcpOptions.val[optionIdx])[3]);
  for (i = 4; i < DhcpOptions.len[optionIdx]; i+=4) {
    fprintf(stdout,",%u.%u.%u.%u",
	    ((unsigned char *)DhcpOptions.val[optionIdx])[i],
	    ((unsigned char *)DhcpOptions.val[optionIdx])[1+i],
	    ((unsigned char *)DhcpOptions.val[optionIdx])[2+i],
	    ((unsigned char *)DhcpOptions.val[optionIdx])[3+i]);
  }
  (void) fprintf(stdout,"\n");
#endif
  
  DhcpOptions.num++;
  return 1;
}

/*****************************************************************************/
int dhcpConfigOptionSetStr(char *optionString, int optionIdx, char *sptr)
{
  int	slen;
  char  *eql;

  if( !optionString || !sptr || (optionIdx < 0) )
    return 0;
#ifdef DEBUG
  (void) fprintf(stderr,"dhcpConfigOptionSetStr(): optionString=%s,optionIdx=%i,sptr=\"%s\"\n",
		 optionString,optionIdx,sptr);
#endif
  eql = sptr;
  eql++;		/* skip over '=' char */
  if( *eql == 39 ) {	/* single quote */
    eql++;
  }
  if( strlen(eql) < 255 )
    slen = strlen(eql) - 2;
  else
    slen = 253;

  if( slen > 0 )
    (void) strncpy(DhcpOptions.val[optionIdx],eql,slen);
#ifdef DEBUG
  (void) fprintf(stdout,"%s=\'%s\'\n",optionString,
		 cleanmetas((char *)DhcpOptions.val[optionIdx]));
#endif
  DhcpOptions.num++;
  return 1;
}

/*****************************************************************************/
int dhcpConfigMsgRecvSet(char *optionString, unsigned char *p, char *sptr)
{
  int	i = 0;
  char  *eql;

  if( !optionString || !sptr || !p )
    return 0;

#ifdef DEBUG
  (void) fprintf(stderr,"dhcpConfigMsgRecvSet(): optionString=%s,px=\"%s\",sptr=\"%s\"\n",
		 optionString,p,sptr);
#endif

  eql = sptr;
  eql++;			/* skip over '=' char */
  for( i = 0; i < 4; i++ ) {
    ((unsigned char *)p)[i] = strtol(eql, (char **)NULL, 10);
    while( (i != 3) && eql && (*eql != '.') ) { eql++; }	 /* Skip over octet */
    eql++;			                         /* skip over period */
  }
#ifdef DEBUG
  (void) fprintf(stdout,"%s=%u.%u.%u.%u\n",optionString,
		 ((unsigned char *)p)[0],
		 ((unsigned char *)p)[1],
		 ((unsigned char *)p)[2],
		 ((unsigned char *)p)[3]);
#endif

  return 1;
}

/*****************************************************************************/
int dhcpConfigHwAddrSet(char *optionString, unsigned char *p, char *sptr)
{
  int	i = 0;
  char  *eql;

  if( !optionString || !sptr || !p )
    return 0;

#ifdef DEBUG
  (void) fprintf(stderr,"dhcpConfigHwAddrSet(): optionString=%s,sptr=\"%s\"\n",
		 optionString,sptr);
#endif
  eql = sptr;
  eql++;			/* skip over '=' char */
  for( i = 0; i < 6; i++ ) {
    ((unsigned char *)p)[i] = strtol(eql, (char **)NULL, 16);
    while( (i != 5) && eql && (*eql != ':') ) { eql++; }	 /* Skip over digit */
    eql++;			                         /* skip over colon */
  }
#ifdef DEBUG
  (void) fprintf(stdout,"%s=%02X:%02X:%02X:%02X:%02X:%02X\n",optionString,
		 ((unsigned char *)p)[0],
		 ((unsigned char *)p)[1],
		 ((unsigned char *)p)[2],
		 ((unsigned char *)p)[3],
		 ((unsigned char *)p)[4],
		 ((unsigned char *)p)[5]);
#endif
  
  return 1;
}

/*****************************************************************************/
int dhcpConfigSetStr(char *optionString, char *p, char *sptr)
{
  int	slen;
  char  *eql;

  if( !optionString || !sptr || !p )
    return 0;
#ifdef DEBUG
  (void) fprintf(stderr,"dhcpConfigSetStr(): optionString=%s,sptr=\"%s\"\n",
		 optionString,sptr);
#endif
  eql = sptr;
  eql++;		/* skip over '=' char */
  if( *eql == 39 ) {	/* single quote */
    eql++;
  }
  if( strlen(eql) < 255 )
    slen = strlen(eql) - 2;
  else
    slen = 253;

  if( slen > 0 )
    (void) strncpy(p,eql,slen);
#ifdef DEBUG
  (void) fprintf(stdout,"%s=\'%s\'\n",optionString,
		 cleanmetas((char *)p));
#endif
  return 1;
}

/*****************************************************************************/
int dhcpConfigRead()
{
  char	*s1;
  char	*s2;
  int	commas;
  int	i = 0;
  char  hostinfo_file[128];
  char	buf[256];
  char  *s;
  char  *eql;
  FILE	*fp;

  memset(DhcpIface.version,0,sizeof(DhcpIface.version));
  strncpy(DhcpIface.version,VERSION,sizeof(DhcpIface.version));
  snprintf(hostinfo_file,sizeof(hostinfo_file),DHCP_HOSTINFO,ConfigDir,IfNameExt);
  syslog(LOG_ERR,"Reading %s\n",hostinfo_file);
  fp=fopen(hostinfo_file,"r");
  if ( fp == (FILE*) 0 ) {
    syslog(LOG_ERR,"dhcpConfigRead: open/read: %m\n");
    return -1;
  }
  /* Have to parse each option, malloc() space for it in DhcpIface.ciaddr or DhcpOptions.val */  

  while( !feof(fp) && ((s=fgets(buf,255,fp)) != (char*) 0) ) {

    if( !strncasecmp(s,"IPADDR",strlen("IPADDR") ) && 
	(eql=strchr(s,'=')) ) {
      eql++;			/* skip over = char */

      for( i = 0; i < 4; i++ ) {
	((unsigned char *)&DhcpIface.ciaddr)[i] = strtol(eql, (char **)NULL, 10);
	while( eql && (*eql != '.') ) { eql++; }	 /* Skip over octet */
	eql++;			                         /* skip over period */
      }

#ifdef DEBUG
      (void) fprintf(stdout,"IPADDR=%u.%u.%u.%u\n",
	      ((unsigned char *)&DhcpIface.ciaddr)[0],
	      ((unsigned char *)&DhcpIface.ciaddr)[1],
	      ((unsigned char *)&DhcpIface.ciaddr)[2],
	      ((unsigned char *)&DhcpIface.ciaddr)[3]);
#endif
      continue;

    }

    if( !strncasecmp(s,"NETMASK",strlen("NETMASK") ) && 
	(eql=strchr(s,'=')) ) {
      if( !(DhcpOptions.val[subnetMask]) ) {
	DhcpOptions.len[subnetMask] = 4;
	DhcpOptions.val[subnetMask] = malloc(4);
      }
      (void) dhcpConfigOptionSet("NETMASK",subnetMask,eql);
      continue;
    }
    if( !strncasecmp(s,"BROADCAST",strlen("BROADCAST") ) && 
	(eql=strchr(s,'=')) ) {
      if( !(DhcpOptions.val[broadcastAddr]) ) {
	DhcpOptions.len[broadcastAddr] = 4;
	DhcpOptions.val[broadcastAddr] = malloc(4);
      }
      (void) dhcpConfigOptionSet("BROADCAST",broadcastAddr,eql);
      continue;
    }

    if( !strncasecmp(s,"GATEWAY",strlen("GATEWAY") ) && 
	(eql=strchr(s,'=')) ) {

      commas = 1;		/* At least one GATEWAY */
      s1 = eql;
      while( (s2 = strchr(s1,',')) ) {
	commas++; /* Count # of gateways */
	s1 = s2 + 1;
      }

      if( !(DhcpOptions.val[routersOnSubnet]) ) {
	DhcpOptions.len[routersOnSubnet] = commas * 4;
	DhcpOptions.val[routersOnSubnet] = malloc(commas * 4);
      }
      (void) dhcpConfigOptionSet("GATEWAY",routersOnSubnet,eql);
      continue;
    }
    if( !strncasecmp(s,"ROUTE",strlen("ROUTE") ) && 
	(eql=strchr(s,'=')) ) {

      commas = 1;		/* At least ONE ROUTE. */
      s1 = eql;
      while( (s2 = strchr(s1,',')) ) {
	commas++; /* Count # of routes */
	s1 = s2 + 1;
      }

      if( !(DhcpOptions.val[staticRoute]) ) {
	DhcpOptions.len[staticRoute] = (commas / 2) * 8;
	DhcpOptions.val[staticRoute] = malloc((commas / 2) * 8);
      }
      (void) dhcpConfigOptionSet("ROUTE",staticRoute,eql);
      continue;
    }
    if( !strncasecmp(s,"HOSTNAME",strlen("HOSTNAME") ) && 
	(eql=strchr(s,'=')) ) {

      if( !(DhcpOptions.val[hostName]) ) {
	DhcpOptions.len[hostName] = 255;
	DhcpOptions.val[hostName] = malloc(255);
	(void) memset(DhcpOptions.val[hostName],0,255);
      }
      (void) dhcpConfigOptionSetStr("HOSTNAME",hostName,eql);
      continue;
    }
    if( !strncasecmp(s,"DOMAIN",strlen("DOMAIN") ) && 
	(eql=strchr(s,'=')) ) {

      if( !(DhcpOptions.val[domainName]) ) {
	DhcpOptions.len[domainName] = 255;
	DhcpOptions.val[domainName] = malloc(255);
	(void) memset(DhcpOptions.val[domainName],0,255);
      }
      (void) dhcpConfigOptionSetStr("DOMAIN",domainName,eql);
      continue;
    }
    if( !strncasecmp(s,"NISDOMAIN",strlen("NISDOMAIN") ) && 
	(eql=strchr(s,'=')) ) {

      if( !(DhcpOptions.val[nisDomainName]) ) {
	DhcpOptions.len[nisDomainName] = 255;
	DhcpOptions.val[nisDomainName] = malloc(255);
	(void) memset(DhcpOptions.val[nisDomainName],0,255);
      }
      (void) dhcpConfigOptionSetStr("NISDOMAIN",nisDomainName,eql);
      continue;
    }
    if( !strncasecmp(s,"ROOTPATH",strlen("ROOTPATH") ) && 
	(eql=strchr(s,'=')) ) {

      if( !(DhcpOptions.val[rootPath]) ) {
	DhcpOptions.len[rootPath] = 255;
	DhcpOptions.val[rootPath] = malloc(255);
	(void) memset(DhcpOptions.val[rootPath],0,255);
      }
      (void) dhcpConfigOptionSetStr("ROOTPATH",rootPath,eql);
      continue;
    }
    if( !strncasecmp(s,"DNS",strlen("DNS") ) && 
	(eql=strchr(s,'=')) ) {

      commas = 1;	/* At least DNS entry */
      s1 = eql;
      while( (s2 = strchr(s1,',')) ) {
	commas++; /* Count # of DNS servers */
	s1 = s2 + 1;
      }

      if( !(DhcpOptions.val[dns]) ) {
	DhcpOptions.len[dns] = commas * 4;
	DhcpOptions.val[dns] = malloc(commas * 4);
      }
      (void) dhcpConfigOptionSet("DNS",dns,eql);
      continue;
    }
    if( !strncasecmp(s,"NTPSERVERS",strlen("NTPSERVERS") ) && 
	(eql=strchr(s,'=')) ) {

      commas = 1;
      s1 = eql;
      while( (s2 = strchr(s1,',')) ) {
	commas++; 
	s1 = s2 + 1;
      }

      if( !(DhcpOptions.val[ntpServers]) ) {
	DhcpOptions.len[ntpServers] = commas * 4;
	DhcpOptions.val[ntpServers] = malloc(commas * 4);
      }
      (void) dhcpConfigOptionSet("NTPSERVERS",ntpServers,eql);
      continue;
    }
    if( !strncasecmp(s,"WINSSERVERS",strlen("WINSSERVERS") ) && 
	(eql=strchr(s,'=')) ) {

      commas = 1;

      s1 = eql;
      while( (s2 = strchr(s1,',')) ) {
	commas++;
	s1 = s2 + 1;
      }

      if( !(DhcpOptions.val[netBIOSnameServer]) ) {
	DhcpOptions.len[netBIOSnameServer] = commas * 4;
	DhcpOptions.val[netBIOSnameServer] = malloc(commas * 4);
      }
      (void) dhcpConfigOptionSet("WINSSERVERS",netBIOSnameServer,eql);
      continue;
    }
    if( !strncasecmp(s,"NISSERVERS",strlen("NISSERVERS") ) && 
	(eql=strchr(s,'=')) ) {

      commas = 1;

      s1 = eql;
      while( (s2 = strchr(s1,',')) ) {
	commas++;
	s1 = s2 + 1;
      }

      if( !(DhcpOptions.val[nisServers]) ) {
	DhcpOptions.len[nisServers] = commas * 4;
	DhcpOptions.val[nisServers] = malloc(commas * 4);
      }
      (void) dhcpConfigOptionSet("NISSERVERS",nisServers,eql);
      continue;
    }
    if( !strncasecmp(s,"LPRSERVERS",strlen("LPRSERVERS") ) && 
	(eql=strchr(s,'=')) ) {

      commas = 1;
      s1 = eql;
      while( (s2 = strchr(s1,',')) ) {
	commas++;
	s1 = s2 + 1;
      }

      if( !(DhcpOptions.val[lprServer]) ) {
	DhcpOptions.len[lprServer] = commas * 4;
	DhcpOptions.val[lprServer] = malloc(commas * 4);
      }
      (void) dhcpConfigOptionSet("LPRSERVERS",lprServer,eql);
      continue;
    }

    if( !strncasecmp(s,"LOGSERVERS",strlen("LOGSERVERS") ) && 
	(eql=strchr(s,'=')) ) {

      commas = 1;
      s1 = eql;
      while( (s2 = strchr(s1,',')) ) {
	commas++;
	s1 = s2 + 1;
      }

      if( !(DhcpOptions.val[logServer]) ) {
	DhcpOptions.len[logServer] = commas * 4;
	DhcpOptions.val[logServer] = malloc(commas * 4);
      }
      (void) dhcpConfigOptionSet("LOGSERVERS",logServer,eql);
      continue;
    }

    if( !strncasecmp(s,"SOHOSERVER",strlen("SOHOSERVER") ) && 
	(eql=strchr(s,'=')) ) {

      commas = 1;
      s1 = eql;
      while( (s2 = strchr(s1,',')) ) {
	commas++;
	s1 = s2 + 1;
      }

      /* if( !(DhcpOptions.val[sohoServer]) ) {	// commented out
	DhcpOptions.len[sohoServer] = commas * 4;
	DhcpOptions.val[sohoServer] = malloc(commas * 4);
      }
      (void) dhcpConfigOptionSet("SOHOSERVER",sohoServer,eql); */
      continue;
    }

    if( !strncasecmp(s,"DHCPSID",strlen("DHCPSID") ) && 
	(eql=strchr(s,'=')) ) {

      if( !(DhcpOptions.val[dhcpServerIdentifier]) ) {
	DhcpOptions.len[dhcpServerIdentifier] = 4;
	DhcpOptions.val[dhcpServerIdentifier] = malloc(4);
      }
      (void) dhcpConfigOptionSet("DHCPSID",dhcpServerIdentifier,eql);
      continue;
    }

    if( !strncasecmp(s,"DHCPGIADDR",strlen("DHCPGIADDR") ) && 
	(eql=strchr(s,'=')) ) {

      (void) dhcpConfigMsgRecvSet("DHCPGIADDR",(unsigned char *)&DhcpMsgRecv->giaddr,eql);
      continue;
    }
    if( !strncasecmp(s,"DHCPSIADDR",strlen("DHCPSIADDR") ) && 
	(eql=strchr(s,'=')) ) {

      (void) dhcpConfigMsgRecvSet("DHCPSIADDR",(unsigned char *)&DhcpMsgRecv->siaddr,eql);
      continue;
    }
    if( !strncasecmp(s,"DHCPCHADDR",strlen("DHCPCHADDR") ) && 
	(eql=strchr(s,'=')) ) {

      (void) dhcpConfigHwAddrSet("DHCPCHADDR",(unsigned char *)&ClientHwAddr,eql);
      continue;
    }
    if( !strncasecmp(s,"DHCPSHADDR",strlen("DHCPSHADDR") ) && 
	(eql=strchr(s,'=')) ) {

      (void) dhcpConfigHwAddrSet("DHCPSHADDR",(unsigned char *)&DhcpIface.shaddr,eql);
      continue;
    }
    if( !strncasecmp(s,"DHCPSNAME",strlen("DHCPSNAME") ) && 
	(eql=strchr(s,'=')) ) {

      (void) memset((void*) &DhcpMsgRecv->sname,0,sizeof(DhcpMsgRecv->sname));
      (void) dhcpConfigSetStr("DHCPSNAME",(char*)DhcpMsgRecv->sname,eql);
      continue;
    }
    if( !strncasecmp(s,"LEASETIME",strlen("LEASETIME") ) && 
	(eql=strchr(s,'=')) ) {

      unsigned	int	ival = 0;

      eql++;			/* Skip over = */
      if( !(DhcpOptions.val[dhcpIPaddrLeaseTime]) ) {
	DhcpOptions.len[dhcpIPaddrLeaseTime] = 4;
	DhcpOptions.val[dhcpIPaddrLeaseTime] = malloc(4);
	DhcpOptions.num++;
      }
      ival = strtoul(eql, (char **)NULL, 10);
      ival = htonl(ival);
      memcpy(DhcpOptions.val[dhcpIPaddrLeaseTime],&ival,4);
#ifdef DEBUG
      (void) fprintf(stdout,"LEASETIME=%u\n", 
		     ntohl(*(unsigned int *)DhcpOptions.val[dhcpIPaddrLeaseTime]));
#endif
      continue;
    }
    if( !strncasecmp(s,"RENEWALTIME",strlen("RENEWALTIME") ) && 
	(eql=strchr(s,'=')) ) {

      unsigned	int	ival = 0;

      eql++;			/* Skip over = */
      if( !(DhcpOptions.val[dhcpT1value]) ) {
	DhcpOptions.len[dhcpT1value] = 4;
	DhcpOptions.val[dhcpT1value] = malloc(4);
	DhcpOptions.num++;
      }
      ival = strtoul(eql, (char **)NULL, 10);
      ival = htonl(ival);
      memcpy(DhcpOptions.val[dhcpT1value],&ival,4);
#ifdef DEBUG
      (void) fprintf(stdout,"RENEWALTIME=%u\n", 
		     ntohl(*(unsigned int *)DhcpOptions.val[dhcpT1value]));
#endif
      continue;
    }
    if( !strncasecmp(s,"REBINDTIME",strlen("REBINDTIME") ) && 
	(eql=strchr(s,'=')) ) {

      unsigned	int	ival = 0;

      eql++;			/* Skip over = */
      if( !(DhcpOptions.val[dhcpT2value]) ) {
	DhcpOptions.len[dhcpT2value] = 4;
	DhcpOptions.val[dhcpT2value] = malloc(4);
	DhcpOptions.num++;
      }
      ival = strtoul(eql, (char **)NULL, 10);
      ival = htonl(ival);
      memcpy(DhcpOptions.val[dhcpT2value],&ival,4);
#ifdef DEBUG
      (void) fprintf(stdout,"REBINDTIME=%u\n", 
		     ntohl(*(unsigned int *)DhcpOptions.val[dhcpT2value]));
#endif
      continue;
    }
    if( !strncasecmp(s,"INTERFACE",strlen("INTERFACE") ) && 
	(eql=strchr(s,'=')) ) {
      static	char  *IfNameExtstr = NULL;

      if( !IfNameExtstr ) {
	IfNameExtstr = (char*) malloc(256);
      }
      (void) memset((void*) IfNameExtstr,0,256);
      (void) dhcpConfigSetStr("INTERFACE",(char*)IfNameExtstr,eql);
      IfNameExt = IfNameExtstr;
#ifdef DEBUG
      (void) fprintf(stdout,"INTERFACE=\'%s\'\n",IfNameExt);
#endif
      continue;
    }
    if( !strncasecmp(s,"CLASSID",strlen("CLASSID") ) && 
	(eql=strchr(s,'=')) ) {

      (void) memset((void*) DhcpIface.class_id,0, sizeof(DhcpIface.class_id));
      (void) dhcpConfigSetStr("CLASSID",(char*)DhcpIface.class_id,eql);
      continue;
    }
    if( !strncasecmp(s,"CLIENTID",strlen("CLIENTID") ) && 
	(eql=strchr(s,'=')) ) {
      unsigned char  mac[6];
      int	     j;

      (void) dhcpConfigHwAddrSet("CLIENTID",(unsigned char *)&mac,eql);
      for(i=0,j=3;i < 6; i++,j++) {
	  DhcpIface.client_id[j] = mac[i];
    }
	  
#ifdef DEBUG
    (void) fprintf(stdout,"CLIENTID=%02X:%02X:%02X:%02X:%02X:%02X\n",
DhcpIface.client_id[3],DhcpIface.client_id[4],DhcpIface.client_id[5],
DhcpIface.client_id[6],DhcpIface.client_id[7],DhcpIface.client_id[8]);
#endif

      continue;
    }

  }
  return 0;
}
/*****************************************************************************/
